from __future__ import annotations

import getpass
import os
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable

from cleo.formatters.style import Style
from cleo.io.outputs.stream_output import StreamOutput


if TYPE_CHECKING:
    from cleo.io.io import IO

Validator = Callable[[str], Any]
Normalizer = Callable[[str], Any]


class Question:
    """
    A question that will be asked in a Console.
    """

    def __init__(self, question: str, default: Any = None) -> None:
        self._question = question
        self._default = default

        self._attempts: int | None = None
        self._hidden = False
        self._hidden_fallback = True
        self._autocomplete_values: list[str] = []
        self._validator: Validator = lambda s: s
        self._normalizer: Normalizer = lambda s: s
        self._error_message = 'Value "{}" is invalid'

    @property
    def question(self) -> str:
        return self._question

    @property
    def default(self) -> Any:
        return self._default

    @property
    def autocomplete_values(self) -> list[str]:
        return self._autocomplete_values

    @property
    def max_attempts(self) -> int | None:
        return self._attempts

    def is_hidden(self) -> bool:
        return self._hidden

    def hide(self, hidden: bool = True) -> None:
        if hidden is True and self._autocomplete_values:
            raise RuntimeError("A hidden question cannot use the autocompleter.")

        self._hidden = hidden

    def set_autocomplete_values(self, autocomplete_values: list[str]) -> None:
        if self.is_hidden():
            raise RuntimeError("A hidden question cannot use the autocompleter.")

        self._autocomplete_values = autocomplete_values

    def set_max_attempts(self, attempts: int | None) -> None:
        self._attempts = attempts

    def set_validator(self, validator: Validator) -> None:
        self._validator = validator

    def ask(self, io: IO) -> Any:
        """
        Asks the question to the user.
        """
        if not io.is_interactive():
            return self.default
        return self._validate_attempts(lambda: self._do_ask(io), io)

    def _do_ask(self, io: IO) -> Any:
        """
        Asks the question to the user.
        """
        self._write_prompt(io)

        if not (self._autocomplete_values and self._has_stty_available()):
            ret: str | None = None

            if self.is_hidden():
                try:
                    ret = self._get_hidden_response(io)
                except RuntimeError:
                    if not self._hidden_fallback:
                        raise

            if not ret:
                ret = self._read_from_input(io)
        else:
            ret = self._autocomplete(io)

        if len(ret) <= 0:
            ret = self._default

        return self._normalizer(ret)  # type: ignore[arg-type]

    def _write_prompt(self, io: IO) -> None:
        """
        Outputs the question prompt.
        """
        io.write_error(f"<question>{self._question}</question> ")

    def _write_error(self, io: IO, error: Exception) -> None:
        """
        Outputs an error message.
        """
        io.write_error_line(f"<error>{error!s}</error>")

    def _autocomplete(self, io: IO) -> str:
        """
        Autocomplete a question.
        """
        autocomplete = self._autocomplete_values

        ret = ""

        i = 0
        ofs = -1
        matches = list(autocomplete)
        num_matches = len(matches)

        # Add highlighted text style
        style = Style(options=["reverse"])
        io.error_output.formatter.set_style("hl", style)

        stty_mode = subprocess.check_output(["stty", "-g"]).decode().rstrip("\n")

        # Disable icanon (so we can read each keypress) and
        # echo (we'll do echoing here instead)
        subprocess.check_output(["stty", "-icanon", "-echo"])
        try:
            # Read a keypress
            while True:
                c = io.read(1)

                # Backspace character
                if c == "\177":
                    if num_matches == 0 and i != 0:
                        i -= 1
                        # Move cursor backwards
                        io.write_error("\033[1D")

                    if i == 0:
                        ofs = -1
                        matches = list(autocomplete)
                        num_matches = len(matches)
                    else:
                        num_matches = 0

                    # Pop the last character off the end of our string
                    ret = ret[:i]
                # Did we read an escape sequence
                elif c == "\033":
                    c += io.read(2)

                    # A = Up Arrow. B = Down Arrow
                    if c[2] == "A" or c[2] == "B":
                        if c[2] == "A" and ofs == -1:
                            ofs = 0

                        if num_matches == 0:
                            continue

                        ofs += -1 if c[2] == "A" else 1
                        ofs = (num_matches + ofs) % num_matches
                elif ord(c) < 32:
                    if c in ["\t", "\n"]:
                        if num_matches > 0 and ofs != -1:
                            ret = matches[ofs]
                            # Echo out remaining chars for current match
                            io.write_error(ret[i:])
                            i = len(ret)

                        if c == "\n":
                            io.write_error(c)
                            break

                        num_matches = 0

                    continue
                else:
                    io.write_error(c)
                    ret += c
                    i += 1

                    num_matches = 0
                    ofs = 0

                    for value in autocomplete:
                        # If typed characters match the beginning
                        # chunk of value (e.g. [AcmeDe]moBundle)
                        if value.startswith(ret) and i != len(value):
                            num_matches += 1
                            matches[num_matches - 1] = value

                # Erase characters from cursor to end of line
                io.write_error("\033[K")

                if num_matches > 0 and ofs != -1:
                    # Save cursor position
                    io.write_error("\0337")
                    # Write highlighted text
                    io.write_error("<hl>" + matches[ofs][i:] + "</hl>")
                    # Restore cursor position
                    io.write_error("\0338")
        finally:
            subprocess.call(["stty", f"{stty_mode}"])

        return ret

    def _get_hidden_response(self, io: IO) -> str:
        """
        Gets a hidden response from user.
        """
        stream = None
        if isinstance(io.error_output, StreamOutput):
            stream = io.error_output.stream
        return getpass.getpass("", stream=stream)

    def _validate_attempts(self, interviewer: Callable[[], Any], io: IO) -> Any:
        """
        Validates an attempt.
        """
        error = None
        attempts = self._attempts

        while attempts is None or attempts:
            if error is not None:
                self._write_error(io, error)

            try:
                return self._validator(interviewer())
            except Exception as e:
                error = e

            if attempts is not None:
                attempts -= 1

        assert error
        raise error

    def _read_from_input(self, io: IO) -> str:
        """
        Read user input.
        """
        ret = io.read_line(4096)

        if not ret:
            raise RuntimeError("Aborted")

        return ret.strip()

    def _has_stty_available(self) -> bool:
        with Path(os.devnull).open("w") as devnull:
            try:
                exit_code = subprocess.call(["stty"], stdout=devnull, stderr=devnull)
            except Exception:
                exit_code = 2

        return exit_code == 0



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from cleo.commands.command import Command


class CommandLoader:
    @property
    def names(self) -> list[str]:
        """
        All registered command names.
        """
        raise NotImplementedError

    def get(self, name: str) -> Command:
        """
        Loads a command.
        """
        raise NotImplementedError

    def has(self, name: str) -> bool:
        """
        Checks whether a command exists or not.
        """
        raise NotImplementedError



from __future__ import annotations

import hashlib
import inspect
import os
import posixpath
import re
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import cast

from cleo import helpers
from cleo._compat import shell_quote
from cleo.commands.command import Command
from cleo.commands.completions.templates import TEMPLATES
from cleo.exceptions import CleoRuntimeError


if TYPE_CHECKING:
    from cleo.io.inputs.argument import Argument
    from cleo.io.inputs.option import Option


class CompletionsCommand(Command):
    name = "completions"
    description = "Generate completion scripts for your shell."

    arguments: ClassVar[list[Argument]] = [
        helpers.argument(
            "shell", "The shell to generate the scripts for.", optional=True
        )
    ]
    options: ClassVar[list[Option]] = [
        helpers.option(
            "alias", None, "Alias for the current command.", flag=False, multiple=True
        )
    ]

    SUPPORTED_SHELLS = ("bash", "zsh", "fish")

    hidden = True

    help = """
One can generate a completion script for `<options=bold>{script_name}</>` \
that is compatible with a given shell. The script is output on \
`<options=bold>stdout</>` allowing one to re-direct \
the output to the file of their choosing. Where you place the file will \
depend on which shell, and which operating system you are using. Your \
particular configuration may also determine where these scripts need \
to be placed.

Here are some common set ups for the three supported shells under \
Unix and similar operating systems (such as GNU/Linux).

<options=bold>BASH</>:

Completion files are commonly stored in `<options=bold>/etc/bash_completion.d/</>`

Run the command:

`<options=bold>{script_name} {command_name} bash >\
 /etc/bash_completion.d/{script_name}.bash-completion</>`

This installs the completion script. You may have to log out and log \
back in to your shell session for the changes to take effect.

<options=bold>FISH</>:

Fish completion files are commonly stored in\
`<options=bold>$HOME/.config/fish/completions</>`

Run the command:

`<options=bold>{script_name} {command_name} fish > \
~/.config/fish/completions/{script_name}.fish</>`

This installs the completion script. You may have to log out and log \
back in to your shell session for the changes to take effect.

<options=bold>ZSH</>:

ZSH completions are commonly stored in any directory listed in your \
`<options=bold>$fpath</>` variable. To use these completions, you must either add the \
generated script to one of those directories, or add your own \
to this list.

Adding a custom directory is often the safest best if you're unsure \
of which directory to use. First create the directory, for this \
example we'll create a hidden directory inside our `<options=bold>$HOME</>` directory

`<options=bold>mkdir ~/.zfunc</>`

Then add the following lines to your `<options=bold>.zshrc</>` \
just before `<options=bold>compinit</>`

`<options=bold>fpath+=~/.zfunc</>`

Now you can install the completions script using the following command

`<options=bold>{script_name} {command_name} zsh > ~/.zfunc/_{script_name}</>`

You must then either log out and log back in, or simply run

`<options=bold>exec zsh</>`

For the new completions to take affect.

<options=bold>CUSTOM LOCATIONS</>:

Alternatively, you could save these files to the place of your choosing, \
such as a custom directory inside your $HOME. Doing so will require you \
to add the proper directives, such as `source`ing inside your login \
script. Consult your shells documentation for how to add such directives.
"""

    def handle(self) -> int:
        shell = self.argument("shell")
        if not shell:
            shell = self.get_shell_type()

        if shell not in self.SUPPORTED_SHELLS:
            raise ValueError(
                f"[shell] argument must be one of {', '.join(self.SUPPORTED_SHELLS)}"
            )

        self.line(self.render(shell))

        return 0

    def render(self, shell: str) -> str:
        if shell == "bash":
            return self.render_bash()
        if shell == "zsh":
            return self.render_zsh()
        if shell == "fish":
            return self.render_fish()

        raise RuntimeError(f"Unrecognized shell: {shell}")

    @staticmethod
    def _get_prog_name_from_stack() -> str:
        package_name = ""
        frame = inspect.currentframe()
        f_back = frame.f_back if frame is not None else None
        f_globals = f_back.f_globals if f_back is not None else None
        # break reference cycle
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frame

        if f_globals is not None:
            package_name = cast(str, f_globals.get("__name__"))

            if package_name == "__main__":
                package_name = cast(str, f_globals.get("__package__"))

            if package_name:
                package_name = package_name.partition(".")[0]

        if not package_name:
            raise CleoRuntimeError("Can not determine package name")

        return package_name

    def _get_script_name_and_path(self) -> tuple[str, str]:
        script_name = self._io.input.script_name or self._get_prog_name_from_stack()
        script_path = posixpath.realpath(script_name)
        script_name = Path(script_path).name

        return script_name, script_path

    def render_bash(self) -> str:
        script_name, script_path = self._get_script_name_and_path()
        aliases = [script_name, script_path, *self.option("alias")]
        function = self._generate_function_name(script_name, script_path)

        # Global options
        assert self.application
        opts = [
            f"--{opt.name}"
            for opt in sorted(self.application.definition.options, key=lambda o: o.name)
        ]

        # Commands + options
        cmds = []
        cmds_opts = []
        for cmd in sorted(self.application.all().values(), key=lambda c: c.name or ""):
            if cmd.hidden or not (cmd.enabled and cmd.name):
                continue
            command_name = shell_quote(cmd.name) if " " in cmd.name else cmd.name
            cmds.append(command_name)
            options = " ".join(
                f"--{opt.name}".replace(":", "\\:")
                for opt in sorted(cmd.definition.options, key=lambda o: o.name)
            )
            cmds_opts += [
                f"            ({command_name})",
                f'            opts="${{opts}} {options}"',
                "            ;;",
                "",  # newline
            ]

        return TEMPLATES["bash"] % {
            "script_name": script_name,
            "function": function,
            "opts": " ".join(opts),
            "cmds": " ".join(cmds),
            "cmds_opts": "\n".join(cmds_opts[:-1]),  # trim trailing newline
            "compdefs": "\n".join(
                f"complete -o default -F {function} {alias}" for alias in aliases
            ),
        }

    def render_zsh(self) -> str:
        script_name, script_path = self._get_script_name_and_path()
        aliases = [script_path, *self.option("alias")]
        function = self._generate_function_name(script_name, script_path)

        def sanitize(s: str) -> str:
            return self._io.output.formatter.remove_format(s)

        # Global options
        assert self.application
        opts = [
            self._zsh_describe(f"--{opt.name}", sanitize(opt.description))
            for opt in sorted(self.application.definition.options, key=lambda o: o.name)
        ]

        # Commands + options
        cmds = []
        cmds_opts = []
        for cmd in sorted(self.application.all().values(), key=lambda c: c.name or ""):
            if cmd.hidden or not (cmd.enabled and cmd.name):
                continue
            command_name = shell_quote(cmd.name) if " " in cmd.name else cmd.name
            cmds.append(self._zsh_describe(command_name, sanitize(cmd.description)))
            options = " ".join(
                self._zsh_describe(f"--{opt.name}", sanitize(opt.description))
                for opt in sorted(cmd.definition.options, key=lambda o: o.name)
            )
            cmds_opts += [
                f"            ({command_name})",
                f"            opts+=({options})",
                "            ;;",
                "",  # newline
            ]

        return TEMPLATES["zsh"] % {
            "script_name": script_name,
            "function": function,
            "opts": " ".join(opts),
            "cmds": " ".join(cmds),
            "cmds_opts": "\n".join(cmds_opts[:-1]),  # trim trailing newline
            "compdefs": "\n".join(f"compdef {function} {alias}" for alias in aliases),
        }

    def render_fish(self) -> str:
        script_name, script_path = self._get_script_name_and_path()
        function = self._generate_function_name(script_name, script_path)

        def sanitize(s: str) -> str:
            return self._io.output.formatter.remove_format(s).replace("'", "\\'")

        # Global options
        assert self.application
        opts = [
            f"complete -c {script_name} -n '__fish{function}_no_subcommand' "
            f"-l {opt.name} -d '{sanitize(opt.description)}'"
            for opt in sorted(self.application.definition.options, key=lambda o: o.name)
        ]

        # Commands + options
        cmds = []
        cmds_opts = []
        namespaces = set()
        for cmd in sorted(self.application.all().values(), key=lambda c: c.name or ""):
            if cmd.hidden or not cmd.enabled or not cmd.name:
                continue
            cmd_path = cmd.name.split(" ")
            namespace = cmd_path[0]
            cmd_name = cmd_path[-1] if " " in cmd.name else cmd.name

            # We either have a command like `poetry add` or a nested (namespaced)
            # command like `poetry cache clear`.
            if len(cmd_path) == 1:
                cmds.append(
                    f"complete -c {script_name} -f -n '__fish{function}_no_subcommand' "
                    f"-a {cmd_name} -d '{sanitize(cmd.description)}'"
                )
                condition = f"__fish_seen_subcommand_from {cmd_name}"
            else:
                # Complete the namespace first
                if namespace not in namespaces:
                    cmds.append(
                        f"complete -c {script_name} -f -n "
                        f"'__fish{function}_no_subcommand' -a {namespace}"
                    )
                # Now complete the command
                subcmds = [
                    name.split(" ")[-1] for name in self.application.all(namespace)
                ]
                cmds.append(
                    f"complete -c {script_name} -f -n '__fish_seen_subcommand_from "
                    f"{namespace}; and not __fish_seen_subcommand_from {' '.join(subcmds)}' "
                    f"-a {cmd_name} -d '{sanitize(cmd.description)}'"
                )
                condition = (
                    f"__fish_seen_subcommand_from {namespace}; "
                    f"and __fish_seen_subcommand_from {cmd_name}"
                )

            cmds_opts += [
                f"# {cmd.name}",
                *[
                    f"complete -c {script_name} "
                    f"-n '{condition}' "
                    f"-l {opt.name} -d '{sanitize(opt.description)}'"
                    for opt in sorted(cmd.definition.options, key=lambda o: o.name)
                ],
                "",  # newline
            ]
            namespaces.add(namespace)

        return TEMPLATES["fish"] % {
            "script_name": script_name,
            "function": function,
            "opts": "\n".join(opts),
            "cmds": "\n".join(cmds),
            "cmds_opts": "\n".join(cmds_opts[:-1]),  # trim trailing newline
            "cmds_names": " ".join(sorted(namespaces)),
        }

    def get_shell_type(self) -> str:
        shell = os.getenv("SHELL")
        if not shell:
            raise RuntimeError(
                "Could not read SHELL environment variable. "
                "Please specify your shell type by passing it as the first argument."
            )

        return Path(shell).name

    def _generate_function_name(self, script_name: str, script_path: str) -> str:
        sanitized_name = self._sanitize_for_function_name(script_name)
        md5_hash = hashlib.md5(script_path.encode()).hexdigest()[:16]
        return f"_{sanitized_name}_{md5_hash}_complete"

    def _sanitize_for_function_name(self, name: str) -> str:
        name = name.replace("-", "_")

        return re.sub(r"[^A-Za-z0-9_]+", "", name)

    def _zsh_describe(self, value: str, description: str | None = None) -> str:
        value = '"' + value.replace(":", "\\:")
        if description:
            description = re.sub(
                r"([\"'#&;`|*?~<>^()\[\]{}$\\\x0A\xFF])", r"\\\1", description
            )
            value += ":" + subprocess.list2cmdline([description]).strip('"')

        value += '"'

        return value



# The COMMAND event allows to attach listeners before any command
# is executed. It also allows the modification of the command and IO
# before it's handed to the command.
from __future__ import annotations


COMMAND = "console.command"

# The SIGNAL event allows some actions to be performed after
# the command execution is interrupted.
SIGNAL = "console.signal"

# The TERMINATE event allows listeners to be attached after the command
# is executed by the console.
TERMINATE = "console.terminate"

# The ERROR event occurs when an uncaught exception is raised.
#
# This event gives the ability to deal with the exception or to modify
# the raised exception.
ERROR = "console.error"



from __future__ import annotations

import math

from dataclasses import dataclass
from html.parser import HTMLParser

from rapidfuzz.distance import Levenshtein


class TagStripper(HTMLParser):
    def __init__(self) -> None:
        super().__init__(convert_charrefs=False)

        self.reset()
        self.fed: list[str] = []

    def handle_data(self, d: str) -> None:
        self.fed.append(d)

    def handle_entityref(self, name: str) -> None:
        self.fed.append(f"&{name};")

    def handle_charref(self, name: str) -> None:
        self.fed.append(f"&#{name};")

    def get_data(self) -> str:
        return "".join(self.fed)


def _strip(value: str) -> str:
    s = TagStripper()
    s.feed(value)
    s.close()

    return s.get_data()


def strip_tags(value: str) -> str:
    while "<" in value and ">" in value:
        new_value = _strip(value)
        if value.count("<") == new_value.count("<"):
            break

        value = new_value

    return value


def find_similar_names(name: str, names: list[str]) -> list[str]:
    """
    Finds names similar to a given command name.
    """
    threshold = 1e3
    distance_by_name = {}

    for actual_name in names:
        # Get Levenshtein distance between the input and each command name
        distance = Levenshtein.distance(name, actual_name)

        is_similar = distance <= len(name) / 3
        substring_index = actual_name.find(name)
        is_substring = substring_index != -1

        if is_similar or is_substring:
            distance_by_name[actual_name] = (
                distance,
                substring_index if is_substring else float("inf"),
            )

    # Only keep results with a distance below the threshold
    distance_by_name = {
        key: value
        for key, value in distance_by_name.items()
        if value[0] < 2 * threshold
    }
    # Display results with shortest distance first
    return sorted(distance_by_name, key=lambda key: distance_by_name[key])


@dataclass
class TimeFormat:
    threshold: int
    alias: str
    divisor: int | None = None

    def apply(self, secs: float) -> str:
        if self.divisor:
            return f"{math.ceil(secs / self.divisor)} {self.alias}"
        return self.alias


_TIME_FORMATS: list[TimeFormat] = [
    TimeFormat(1, "< 1 sec"),
    TimeFormat(2, "1 sec"),
    TimeFormat(60, "secs", 1),
    TimeFormat(61, "1 min"),
    TimeFormat(3600, "mins", 60),
    TimeFormat(5401, "1 hr"),
    TimeFormat(86400, "hrs", 3600),
    TimeFormat(129601, "1 day"),
    TimeFormat(604801, "days", 86400),
]


def format_time(secs: float) -> str:
    time_format = next(
        (fmt for fmt in _TIME_FORMATS if secs < fmt.threshold), _TIME_FORMATS[-1]
    )
    return time_format.apply(secs)



from __future__ import annotations

import math
import re
import time

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import Match

from cleo._utils import format_time
from cleo.cursor import Cursor
from cleo.io.io import IO
from cleo.io.outputs.section_output import SectionOutput
from cleo.terminal import Terminal
from cleo.ui.component import Component


if TYPE_CHECKING:
    from cleo.io.outputs.output import Output


class ProgressBar(Component):
    """
    The ProgressBar provides helpers to display progress output.
    """

    name = "progress_bar"

    # Options
    bar_width = 28
    bar_char = None
    empty_bar_char = "-"
    progress_char = ">"
    redraw_freq: int | None = 1

    formats: ClassVar[dict[str, str]] = {
        "normal": " %current%/%max% [%bar%] %percent:3s%%",
        "normal_nomax": " %current% [%bar%]",
        "verbose": " %current%/%max% [%bar%] %percent:3s%% %elapsed:-6s%",
        "verbose_nomax": " %current% [%bar%] %elapsed:6s%",
        "very_verbose": (
            " %current%/%max% [%bar%] %percent:3s%%" " %elapsed:6s%/%estimated:-6s%"
        ),
        "very_verbose_nomax": " %current% [%bar%] %elapsed:6s%",
        "debug": " %current%/%max% [%bar%] %percent:3s%% %elapsed:6s%/%estimated:-6s%",
        "debug_nomax": " %current% [%bar%] %elapsed:6s%",
    }

    def __init__(
        self,
        io: IO | Output,
        max: int = 0,
        min_seconds_between_redraws: float = 0.1,
    ) -> None:
        # If we have an IO, ensure we write to the error output
        if isinstance(io, IO):
            io = io.error_output

        self._io = io
        self._terminal = Terminal().size
        self._max = 0
        self._step_width: int = 1
        self._set_max_steps(max)
        self._step = 0
        self._percent = 0.0
        self._format: str | None = None
        self._internal_format: str | None = None
        self._format_line_count = 0
        self._previous_message: str | None = None
        self._should_overwrite = True
        self._min_seconds_between_redraws = 0.0
        self._max_seconds_between_redraws = 1.0
        self._write_count = 0

        if min_seconds_between_redraws > 0:
            self.redraw_freq = None
            self._min_seconds_between_redraws = min_seconds_between_redraws

        if not self._io.formatter.is_decorated():
            # Disable overwrite when output does not support ANSI codes.
            self._should_overwrite = False

            # Set a reasonable redraw frequency so output isn't flooded
            self.redraw_freq = None

        self._messages: dict[str, str] = {}

        self._start_time = time.time()
        self._last_write_time = 0.0
        self._cursor = Cursor(self._io)

    def set_message(self, message: str, name: str = "message") -> None:
        self._messages[name] = message

    def get_message(self, name: str = "message") -> str:
        return self._messages[name]

    def get_start_time(self) -> float:
        return self._start_time

    def get_max_steps(self) -> int:
        return self._max

    def get_progress(self) -> int:
        return self._step

    def get_progress_percent(self) -> float:
        return self._percent

    def set_bar_character(self, character: str) -> ProgressBar:
        self.bar_char = character

        return self

    def get_bar_character(self) -> str:
        if self.bar_char is None:
            if self._max:
                return "="

            return self.empty_bar_char

        return self.bar_char

    def set_bar_width(self, width: int) -> ProgressBar:
        self.bar_width = width

        return self

    def get_empty_bar_character(self) -> str:
        return self.empty_bar_char

    def set_empty_bar_character(self, character: str) -> ProgressBar:
        self.empty_bar_char = character

        return self

    def get_progress_character(self) -> str:
        return self.progress_char

    def set_progress_character(self, character: str) -> ProgressBar:
        self.progress_char = character

        return self

    def set_format(self, fmt: str) -> None:
        self._format = None
        self._internal_format = fmt

    def set_redraw_frequency(self, freq: int) -> None:
        if self.redraw_freq is not None:
            self.redraw_freq = max(freq, 1)

    def min_seconds_between_redraws(self, freq: float) -> None:
        if freq > 0:
            self.redraw_freq = None
            self._min_seconds_between_redraws = freq

    def max_seconds_between_redraws(self, freq: float) -> None:
        self._max_seconds_between_redraws = freq

    def start(self, max: int | None = None) -> None:
        """
        Start the progress output.
        """
        self._start_time = time.time()
        self._step = 0
        self._percent = 0.0

        if max is not None:
            self._set_max_steps(max)

        self.display()

    def advance(self, step: int = 1) -> None:
        """
        Advances the progress output X steps.
        """
        self.set_progress(self._step + step)

    def set_progress(self, step: int) -> None:
        """
        Sets the current progress.
        """
        if self._max and step > self._max:
            self._max = step
        elif step < 0:
            step = 0

        redraw_freq = (
            (self._max or 10) / 10 if self.redraw_freq is None else self.redraw_freq
        )
        prev_period = int(self._step / redraw_freq)
        curr_period = int(step / redraw_freq)

        self._step = step
        self._percent = step / (self._max or math.inf)

        time_interval = time.time() - self._last_write_time

        # Draw regardless of other limits
        if step == self._max:
            self.display()

            return

        # Throttling
        if time_interval < self._min_seconds_between_redraws:
            return

        # Draw each step period, but not too late
        if (
            prev_period != curr_period
            or time_interval >= self._max_seconds_between_redraws
        ):
            self.display()

    def finish(self) -> None:
        """
        Finish the progress output.
        """
        if not self._max:
            self._max = self._step

        if self._step == self._max and not self._should_overwrite:
            return

        self.set_progress(self._max)

    def display(self) -> None:
        """
        Output the current progress string.
        """
        if self._io.is_quiet():
            return

        if self._format is None:
            self._set_real_format(
                self._internal_format or self._determine_best_format()
            )

        self._overwrite(self._build_line())

    def _overwrite_callback(self, matches: Match[str]) -> str:
        if hasattr(self, f"_formatter_{matches.group(1)}"):
            text = str(getattr(self, f"_formatter_{matches.group(1)}")())
        elif matches.group(1) in self._messages:
            text = self._messages[matches.group(1)]
        else:
            return matches.group(0)

        if matches.group(2):
            n = int(matches.group(2).lstrip("-").rstrip("s"))
            if matches.group(2).startswith("-"):
                return text.ljust(n)
            return text.rjust(n)

        return text

    def clear(self) -> None:
        """
        Removes the progress bar from the current line.

        This is useful if you wish to write some output
        while a progress bar is running.
        Call display() to show the progress bar again.
        """
        if not self._should_overwrite:
            return

        if self._format is None:
            self._set_real_format(
                self._internal_format or self._determine_best_format()
            )

        self._overwrite("\n" * self._format_line_count)

    def _set_real_format(self, fmt: str) -> None:
        """
        Sets the progress bar format.
        """
        # try to use the _nomax variant if available
        if not self._max and fmt + "_nomax" in self.formats:
            self._format = self.formats[fmt + "_nomax"]
        else:
            self._format = self.formats.get(fmt, fmt)
        assert self._format is not None
        self._format_line_count = self._format.count("\n")

    def _set_max_steps(self, mx: int) -> None:
        """
        Sets the progress bar maximal steps.
        """
        self._max = max(0, mx)
        self._step_width = len(str(self._max)) if self._max else 4

    def _overwrite(self, message: str) -> None:
        """
        Overwrites a previous message to the output.
        """
        if self._previous_message == message:
            return

        original_message = message

        if self._should_overwrite:
            if self._previous_message is not None:
                if isinstance(self._io, SectionOutput):
                    lines_to_clear = (
                        len(self._io.remove_format(message)) // self._terminal.width
                        + self._format_line_count
                        + 1
                    )
                    self._io.clear(lines_to_clear)
                else:
                    if self._format_line_count:
                        self._cursor.move_up(self._format_line_count)

                    self._cursor.move_to_column(1)
                    self._cursor.clear_line()
        elif self._step > 0:
            message = "\n" + message

        self._previous_message = original_message
        self._last_write_time = time.time()

        self._io.write(message)
        self._write_count += 1

    def _determine_best_format(self) -> str:
        fmt = "normal"
        if self._io.is_debug():
            fmt = "debug"
        elif self._io.is_very_verbose():
            fmt = "very_verbose"
        elif self._io.is_verbose():
            fmt = "verbose"

        return fmt if self._max else f"{fmt}_nomax"

    @property
    def bar_offset(self) -> int:
        if self._max:
            return math.floor(self._percent * self.bar_width)
        if self.redraw_freq is None:
            return math.floor(
                (min(5, self.bar_width // 15) * self._write_count) % self.bar_width
            )
        return math.floor(self._step % self.bar_width)

    def _formatter_bar(self) -> str:
        complete_bars = self.bar_offset

        display = self.get_bar_character() * int(complete_bars)

        if complete_bars < self.bar_width:
            empty_bars = (
                self.bar_width
                - complete_bars
                - len(self._io.remove_format(self.progress_char))
            )
            display += self.progress_char + self.empty_bar_char * int(empty_bars)

        return display

    def _formatter_elapsed(self) -> str:
        return format_time(time.time() - self._start_time)

    def _formatter_remaining(self) -> str:
        if not self._max:
            raise RuntimeError(
                "Unable to display the remaining time "
                "if the maximum number of steps is not set."
            )

        if not self._step:
            remaining = 0
        else:
            remaining = round(
                (time.time() - self._start_time) / self._step * (self._max - self._max)
            )

        return format_time(remaining)

    def _formatter_estimated(self) -> int:
        if not self._max:
            raise RuntimeError(
                "Unable to display the estimated time "
                "if the maximum number of steps is not set."
            )

        if not self._step:
            return 0

        return round((time.time() - self._start_time) / self._step * self._max)

    def _formatter_current(self) -> str:
        return str(self._step).rjust(self._step_width)

    def _formatter_max(self) -> int:
        return self._max

    def _formatter_percent(self) -> int:
        return int(math.floor(self._percent * 100))

    def _build_line(self) -> str:
        regex = re.compile(r"(?i)%([a-z\-_]+)(?::([^%]+))?%")
        assert self._format is not None
        line = regex.sub(self._overwrite_callback, self._format)

        # gets string length for each sub line with multiline format
        lines_length = [
            len(self._io.remove_format(sub_line.rstrip("\r")))
            for sub_line in line.split("\n")
        ]

        lines_width = max(lines_length)

        terminal_width = self._terminal.width

        if lines_width <= terminal_width:
            return line

        self.set_bar_width(self.bar_width - lines_width + terminal_width)

        return regex.sub(self._overwrite_callback, self._format)



from __future__ import annotations

import re

from typing import TYPE_CHECKING
from typing import Any
from typing import cast

from cleo.exceptions import CleoValueError
from cleo.ui.question import Question


if TYPE_CHECKING:
    from cleo.io.io import IO


class SelectChoiceValidator:
    def __init__(self, question: ChoiceQuestion) -> None:
        """
        Constructor.
        """
        self._question = question
        self._values = question.choices

    def validate(self, selected: Any) -> str | list[str] | None:
        """
        Validate a choice.
        """
        # Collapse all spaces.
        if isinstance(selected, int):
            selected = str(selected)

        if selected is None:
            return None

        if self._question.supports_multiple_choices():
            # Check for a separated comma values
            _selected = selected.replace(" ", "")
            if not re.match(r"^[a-zA-Z0-9_-]+(?:,[a-zA-Z0-9_-]+)*$", _selected):
                raise CleoValueError(self._question.error_message.format(selected))

            selected_choices = _selected.split(",")
        else:
            selected_choices = [selected]

        multiselect_choices = []
        for value in selected_choices:
            results = []

            for key, choice in enumerate(self._values):
                if choice == value:
                    results.append(key)

            if len(results) > 1:
                raise CleoValueError(
                    "The provided answer is ambiguous. "
                    f"Value should be one of {' or '.join(str(r) for r in results)}."
                )

            if value in self._values:
                result = value
            elif value.isdigit() and 0 <= int(value) < len(self._values):
                result = self._values[int(value)]
            else:
                raise CleoValueError(self._question.error_message.format(value))

            multiselect_choices.append(result)

        if self._question.supports_multiple_choices():
            return multiselect_choices

        return cast("str | list[str] | None", multiselect_choices[0])


class ChoiceQuestion(Question):
    """
    Multiple choice question.
    """

    def __init__(
        self, question: str, choices: list[str], default: Any | None = None
    ) -> None:
        super().__init__(question, default)

        self._multi_select = False
        self._choices = choices
        self._validator = SelectChoiceValidator(self).validate
        self._autocomplete_values = choices
        self._prompt = " > "
        self._error_message = 'Value "{}" is invalid'

    @property
    def error_message(self) -> str:
        return self._error_message

    @property
    def choices(self) -> list[str]:
        return self._choices

    def supports_multiple_choices(self) -> bool:
        return self._multi_select

    def set_multi_select(self, multi_select: bool) -> None:
        self._multi_select = multi_select

    def set_error_message(self, message: str) -> None:
        self._error_message = message

    def _write_prompt(self, io: IO) -> None:
        """
        Outputs the question prompt.
        """
        message = self._question
        default = self._default

        if default is None:
            message = f"<question>{message}</question>: "
        elif self._multi_select:
            choices = self._choices
            default = default.split(",")

            for i, value in enumerate(default):
                default[i] = choices[int(value.strip())]

            message = (
                f"<question>{message}</question> "
                f"[<comment>{', '.join(default)}</comment>]:"
            )
        else:
            choices = self._choices
            message = (
                f"<question>{message}</question> "
                f"[<comment>{choices[int(default)]}</comment>]:"
            )

        width = len(str(len(self._choices) - 1)) if len(self._choices) > 1 else 1

        messages = [message]
        for key, value in enumerate(self._choices):
            messages.append(f" [<comment>{key: {width}}</>] {value}")

        io.write_error_line("\n".join(messages))

        message = self._prompt

        io.write_error(message)



from __future__ import annotations

import re

from typing import ClassVar

from cleo.exceptions import CleoValueError
from cleo.formatters.style import Style
from cleo.formatters.style_stack import StyleStack


class Formatter:
    TAG_REGEX = re.compile(r"(?ix)<(([a-z](?:[^<>]*)) | /([a-z](?:[^<>]*))?)>")

    _inline_styles_cache: ClassVar[dict[str, Style]] = {}

    def __init__(
        self, decorated: bool = False, styles: dict[str, Style] | None = None
    ) -> None:
        self._decorated = decorated
        self._styles: dict[str, Style] = {}

        self.set_style("error", Style("red", options=["bold"]))
        self.set_style("info", Style("blue"))
        self.set_style("comment", Style("green"))
        self.set_style("question", Style("cyan"))
        self.set_style("c1", Style("cyan"))
        self.set_style("c2", Style("default", options=["bold"]))
        self.set_style("b", Style("default", options=["bold"]))

        for name, style in (styles or {}).items():
            self.set_style(name, style)

        self._style_stack = StyleStack()

    @classmethod
    def escape(cls, text: str) -> str:
        """
        Escapes "<" special char in given text.
        """
        text = re.sub(r"([^\\]?)<", "\\1\\<", text)

        return cls.escape_trailing_backslash(text)

    @staticmethod
    def escape_trailing_backslash(text: str) -> str:
        """
        Escapes trailing "\\" in given text.
        """
        if text.endswith("\\"):
            length = len(text)
            text = text.rstrip("\\").replace("\0", "").ljust(length, "\0")

        return text

    def decorated(self, decorated: bool = True) -> None:
        self._decorated = decorated

    def is_decorated(self) -> bool:
        return self._decorated

    def set_style(self, name: str, style: Style) -> None:
        self._styles[name] = style

    def has_style(self, name: str) -> bool:
        return name in self._styles

    def style(self, name: str) -> Style:
        if not self.has_style(name):
            raise CleoValueError(f'Undefined style: "{name}"')

        return self._styles[name]

    def format(self, message: str) -> str:
        return self.format_and_wrap(message, 0)

    def format_and_wrap(self, message: str, width: int) -> str:
        offset = 0
        output = ""
        current_line_length = 0
        for match in self.TAG_REGEX.finditer(message):
            pos = match.start()
            text = match.group(0)

            if pos != 0 and message[pos - 1] == "\\":
                continue

            # add the text up to the next tag
            formatted, current_line_length = self._apply_current_style(
                message[offset:pos], output, width, current_line_length
            )
            output += formatted
            offset = pos + len(text)

            # Opening tag
            seen_open = text[1] != "/"
            tag = match.group(1) if seen_open else match.group(2)

            style = None
            if tag:
                style = self._create_style_from_string(tag)

            if not (seen_open or tag):
                # </>
                self._style_stack.pop()
            elif style is None:
                formatted, current_line_length = self._apply_current_style(
                    text, output, width, current_line_length
                )
                output += formatted
            elif seen_open:
                self._style_stack.push(style)
            else:
                self._style_stack.pop(style)

        formatted, current_line_length = self._apply_current_style(
            message[offset:], output, width, current_line_length
        )
        output += formatted
        return output.replace("\0", "\\").replace("\\<", "<")

    def remove_format(self, text: str) -> str:
        decorated = self._decorated

        self._decorated = False
        text = re.sub(r"\033\[[^m]*m", "", self.format(text))

        self._decorated = decorated

        return text

    def _create_style_from_string(self, string: str) -> Style | None:
        if string in self._styles:
            return self._styles[string]

        if string in self._inline_styles_cache:
            return self._inline_styles_cache[string]

        matches = re.findall(r"([^=]+)=([^;]+)(;|$)", string.lower())
        if not matches:
            return None

        style = Style()

        for where, style_options, _ in matches:
            if where == "fg":
                style.foreground(style_options)
            elif where == "bg":
                style.background(style_options)
            else:
                try:
                    for option in map(str.strip, style_options.split(",")):
                        style.set_option(option)
                except ValueError:
                    return None

        self._inline_styles_cache[string] = style

        return style

    def _apply_current_style(
        self, text: str, current: str, width: int, current_line_length: int
    ) -> tuple[str, int]:
        if not text:
            return "", current_line_length

        if not width:
            if self.is_decorated():
                return self._style_stack.current.apply(text), current_line_length

            return text, current_line_length

        if not current_line_length and current:
            text = text.lstrip()

        if current_line_length:
            i = width - current_line_length
            prefix = text[:i] + "\n"
            text = text[i:]
        else:
            prefix = ""

        m = re.match(r"(\n)$", text)
        text = prefix + re.sub(rf"([^\n]{{{width}}})\ *", "\\1\n", text)
        text = text.rstrip("\n") + (m.group(1) if m else "")

        if not current_line_length and current and not current.endswith("\n"):
            text = "\n" + text

        lines = text.split("\n")
        for line in lines:
            current_line_length += len(line)
            if current_line_length >= width:
                current_line_length = 0

        if self.is_decorated():
            apply = self._style_stack.current.apply
            text = "\n".join(map(apply, lines))

        return text, current_line_length



from __future__ import annotations

import os
import sys

from typing import NamedTuple


class TerminalSize(NamedTuple):
    width: int
    height: int


class Terminal:
    def __init__(
        self,
        width: int | None = None,
        height: int | None = None,
        fallback: tuple[int, int] | None = None,
    ) -> None:
        self._width = width
        self._height = height
        self._fallback = TerminalSize(*(fallback or (80, 25)))

    @property
    def width(self) -> int:
        return self.size.width

    @property
    def height(self) -> int:
        return self.size.height

    @property
    def size(self) -> TerminalSize:
        return self._get_terminal_size()

    def _get_terminal_size(self) -> TerminalSize:
        if not (self._width is None or self._height is None):
            return TerminalSize(self._width, self._height)

        width = 0
        height = 0

        columns = os.environ.get("COLUMNS")
        if columns is not None and columns.isdigit():
            width = int(columns)
        lines = os.environ.get("LINES")
        if lines is not None and lines.isdigit():
            height = int(lines)

        if width <= 0 or height <= 0:
            try:
                os_size = os.get_terminal_size(sys.__stdout__.fileno())
                size = TerminalSize(*os_size)
            except (AttributeError, ValueError, OSError):
                # stdout is None, closed, detached, or not a terminal, or
                # os.get_terminal_size() is unsupported # noqa: ERA001
                size = self._fallback
            if width <= 0:
                width = size.width or self._fallback.width
            if height <= 0:
                height = size.height or self._fallback.height

        return TerminalSize(
            width if self._width is None else self._width,
            height if self._height is None else self._height,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from cleo.events.console_event import ConsoleEvent
from cleo.exceptions import CleoError


if TYPE_CHECKING:
    from cleo.commands.command import Command
    from cleo.io.io import IO


class ConsoleErrorEvent(ConsoleEvent):
    """
    An event triggered when an exception is raised during the execution of a command.
    """

    def __init__(self, command: Command, io: IO, error: Exception) -> None:
        super().__init__(command, io)

        self._error = error
        self._exit_code: int | None = None

    @property
    def error(self) -> Exception:
        return self._error

    @property
    def exit_code(self) -> int:
        if self._exit_code is not None:
            return self._exit_code

        if isinstance(self._error, CleoError) and self._error.exit_code is not None:
            return self._error.exit_code

        return 1

    def set_error(self, error: Exception) -> None:
        self._error = error

    def set_exit_code(self, exit_code: int) -> None:
        self._exit_code = exit_code



from __future__ import annotations

from cleo.exceptions import CleoValueError
from cleo.ui.component import Component


class UI:
    def __init__(self, components: list[Component] | None = None) -> None:
        self._components: dict[str, Component] = {}

        for component in components or []:
            self.register(component)

    def register(self, component: Component) -> None:
        if not isinstance(component, Component):
            raise CleoValueError(
                "A UI component must inherit from the Component class."
            )

        if not component.name:
            raise CleoValueError("A UI component cannot be anonymous.")

        self._components[component.name] = component

    def component(self, name: str) -> Component:
        if name not in self._components:
            raise CleoValueError(f'UI component "{name}" does not exist.')

        return self._components[name]



from __future__ import annotations

from cleo.exceptions import CleoValueError
from cleo.formatters.style import Style


class StyleStack:
    def __init__(self, empty_style: Style | None = None) -> None:
        if empty_style is None:
            empty_style = Style()

        self._empty_style = empty_style
        self._styles: list[Style] = []

    @property
    def current(self) -> Style:
        if not self._styles:
            return self._empty_style

        return self._styles[-1]

    def reset(self) -> None:
        self._styles = []

    def push(self, style: Style) -> None:
        self._styles.append(style)

    def pop(self, style: Style | None = None) -> Style:
        if not self._styles:
            return self._empty_style

        if style is None:
            return self._styles.pop()

        sample = style.apply("")

        for i, stacked_style in reversed(list(enumerate(self._styles))):
            if sample == stacked_style.apply(""):
                self._styles = self._styles[:i]
                return stacked_style

        raise CleoValueError("Invalid nested tag found")



from __future__ import annotations

import re

from typing import Any
from typing import TextIO

from cleo._compat import shell_quote
from cleo.exceptions import CleoMissingArgumentsError
from cleo.exceptions import CleoValueError
from cleo.io.inputs.definition import Definition


class Input:
    """
    This class is the base class for concrete Input implementations.
    """

    def __init__(self, definition: Definition | None = None) -> None:
        self._definition: Definition
        self._stream: TextIO = None  # type: ignore[assignment]
        self._options: dict[str, Any] = {}
        self._arguments: dict[str, Any] = {}
        self._interactive: bool | None = None

        if definition is None:
            self._definition = Definition()
        else:
            self.bind(definition)
            self.validate()

    @property
    def arguments(self) -> dict[str, Any]:
        return {**self._definition.argument_defaults, **self._arguments}

    @property
    def options(self) -> dict[str, Any]:
        return {**self._definition.option_defaults, **self._options}

    @property
    def stream(self) -> TextIO:
        return self._stream

    @property
    def first_argument(self) -> str | None:
        """
        Returns the first argument from the raw parameters (not parsed).
        """
        raise NotImplementedError

    @property
    def script_name(self) -> str | None:
        raise NotImplementedError

    def read(self, length: int, default: str = "") -> str:
        """
        Reads the given amount of characters from the input stream.
        """
        if not self.is_interactive():
            return default

        return self._stream.read(length)

    def read_line(self, length: int = -1, default: str = "") -> str:
        """
        Reads a line from the input stream.
        """
        if not self.is_interactive():
            return default

        return self._stream.readline(length)

    def close(self) -> None:
        """
        Closes the input.
        """
        self._stream.close()

    def is_closed(self) -> bool:
        """
        Returns whether the input is closed.
        """
        return self._stream.closed

    def is_interactive(self) -> bool:
        return True if self._interactive is None else self._interactive

    def interactive(self, interactive: bool = True) -> None:
        self._interactive = interactive

    def bind(self, definition: Definition) -> None:
        """
        Binds the current Input instance with
        the given definition's arguments and options.
        """
        self._arguments = {}
        self._options = {}
        self._definition = definition

        self._parse()

    def validate(self) -> None:
        missing_arguments = []

        for argument in self._definition.arguments:
            if argument.name not in self._arguments and argument.is_required():
                missing_arguments.append(argument.name)

        if missing_arguments:
            raise CleoMissingArgumentsError(
                f'Not enough arguments (missing: "{", ".join(missing_arguments)}")'
            )

    def argument(self, name: str) -> Any:
        if not self._definition.has_argument(name):
            raise CleoValueError(f'The argument "{name}" does not exist')

        if name in self._arguments:
            return self._arguments[name]

        return self._definition.argument(name).default

    def set_argument(self, name: str, value: Any) -> None:
        if not self._definition.has_argument(name):
            raise CleoValueError(f'The argument "{name}" does not exist')

        self._arguments[name] = value

    def has_argument(self, name: str) -> bool:
        return self._definition.has_argument(name)

    def option(self, name: str) -> Any:
        if not self._definition.has_option(name):
            raise CleoValueError(f'The option "--{name}" does not exist')

        if name in self._options:
            return self._options[name]

        return self._definition.option(name).default

    def set_option(self, name: str, value: Any) -> None:
        if not self._definition.has_option(name):
            raise CleoValueError(f'The option "--{name}" does not exist')

        self._options[name] = value

    def has_option(self, name: str) -> bool:
        return self._definition.has_option(name)

    def escape_token(self, token: str) -> str:
        if re.match(r"^[\w-]+$", token):
            return token

        return shell_quote(token)

    def set_stream(self, stream: TextIO) -> None:
        self._stream = stream

    def has_parameter_option(
        self, values: str | list[str], only_params: bool = False
    ) -> bool:
        """
        Returns true if the raw parameters (not parsed) contain a value.
        """
        raise NotImplementedError

    def parameter_option(
        self,
        values: str | list[str],
        default: Any = False,
        only_params: bool = False,
    ) -> Any:
        """
        Returns the value of a raw option (not parsed).
        """
        raise NotImplementedError

    def _parse(self) -> None:
        raise NotImplementedError
